(* file: main.sml *)
(* description: checker of tree automaton *)
(* author: Takahito Aoto *)

signature MAIN = 
sig
    val toolname: string
    val version: string
    val helpMsg: string
    val compile: string -> string list * string list * ((string * string) * string list) list * string list * string list
    val main: string * string list -> OS.Process.status
end

structure Main = 
struct

local 
    structure L = List
    structure U = Util
    structure S = String
in

val toolname = "checkmyfa"
val version = "0.01"
val helpMsg = 
    toolname ^ " v" ^ version ^ "\n"
    ^ "Usage: sml @SMLload=" ^ toolname ^ ".x86-linux directive\n"
    ^ "directives  --dfa-check filename\n"
    ^ "            --dfa-accept string [--verbose] filename\n"
    ^ "            --dfa-complement filename\n"
    ^ "            --dfa-union filename1 filename2\n"
    ^ "            --dfa-product filename1 filename2\n"
    ^ "            --dfa-empty filename\n"
    ^ "            --dfa-subset filename1 filename2\n"
    ^ "            --dfa-equiv filename1 filename2\n"
    ^ "            --dfa-minimize filename\n"
    ^ "            --nfa-check filename\n"
    ^ "            --nfa-accept string [--verbose] filename\n"
    ^ "            --nfa-concat filename1 filename2\n"
    ^ "            --nfa-union filename1 filename2\n"
    ^ "            --nfa-star filename\n"
    ^ "            --nfa-collapse filename\n"
    ^ "            --nfa-mkdfa filename\n"
    ^ "            --rename [--prefix=string] [--index=num] filename\n"
    ^ "            --re-simplify regexpr\n"
    ^ "            --re-mknfa alphabet regexpr\n"
    ^ "            --dfa-mkre [--verbose] filename\n"
    ^ "            --dfa-pump string filename\n"

structure TpdbLrVals = TpdbLrValsFun (structure Token = LrParser.Token);
structure TpdbLex = TpdbLexFun (structure Tokens = TpdbLrVals.Tokens);
structure TpdbParser = Join (structure ParserData = TpdbLrVals.ParserData;
                             structure Lex = TpdbLex;
                             structure LrParser = LrParser);

exception Error of string (* fatal error *)
fun printError msg = TextIO.output (TextIO.stdErr, "Error: " ^ msg ^ "\n")

fun compile istream =
    let val lexer = TpdbParser.makeLexer (fn n=>TextIO.inputN (istream, n));
	fun error (msg,line,col) = raise Error (msg ^ " at line " ^ (Int.toString line) ^ ", char "^ Int.toString col)
    in #1 (TpdbParser.parse (15, lexer, error, ()))
    end

fun compileFile filename = compile (TextIO.openIn filename)
			   handle Io => raise Error ("failed to read the file " ^ filename)

fun compileStdIn () = compile TextIO.stdIn

val invalidAlphabetChars = [#"+", #"*", #".", #"(", #")",#"<", #">", #"|", #"\"" ]
fun isAlphabetChar c  = not (U.member c invalidAlphabetChars)
fun isEnclosedChar c = Char.isAlphaNum c orelse c = #"-" (* <...> に現われていよい文字 *)
fun isValidSymbol str = let val cs = String.explode str
			in (length cs = 1 andalso isAlphabetChar (hd cs))
			   orelse 
			   ((length cs > 1) andalso ((hd cs) = #"<")
			    andalso let val cs2 = rev cs
				    in ((hd cs2) = #">") andalso L.all isEnclosedChar (tl cs2)
				    end)
			end

(* reading the input string for *)
fun readInput str = 
    let fun read ss = let val (lhs,rhs) = Substring.splitl (fn c => c <> #"<") ss
			  val heads0 = Substring.explode lhs
			  val _ = if List.all isAlphabetChar heads0
				  then ()
				  else raise Error "invalid input string"
			  val heads = List.map String.str heads0
		      in if Substring.size rhs <> 0
			 then let val (name0,rest0) = Substring.splitl (fn c => c <> #">") rhs
				  val name = (Substring.string name0) ^ ">"
			   val _ = if List.all isEnclosedChar (Substring.explode (Substring.triml 1 name0))
				   then () else raise Error "invalid input string"
			   val _ = case Substring.first rest0 of 
				       SOME #">" => ()
				     | _ => raise Error "invalid input string"
			   val rest = Substring.triml 1 rest0
		       in heads @ [name] @ (read rest)
		       end
		  else heads
	      end
    in L.filter (fn s => s <> "<EPSILON>") (read (Substring.full str))
    end

(* common check for DFA and NFA *)
fun isValidSpec (states,alphabet,transitions,start,accept) =
    let 
	val _ = case U.duplicatingElem states of
		    NONE => ()
		  | SOME q => raise Error ("state " ^ q ^ " is multiply presented in states")
	val _ = case U.duplicatingElem alphabet of
		    NONE => ()
		  | SOME x => raise Error ("symbol " ^ x ^ " is multiply presented in alphabet")
	val _ = case U.duplicatingElem accept of
		    NONE => ()
		  | SOME x => raise Error ("symbol " ^ x ^ " is multiply presented in accept states")
	val _ = if length start <> 1
		then raise Error "multiple start states"
		else ()
	val _ = if U.subseteq (start, states)
		then ()
		else raise Error "start states not involved in the set of states"
	val _ = if U.subseteq (accept,states)
		then ()
		else raise Error "accept states not included in the set of states"
	val _ = if List.all (fn str=> ((String.size str = 1) 
				       andalso isAlphabetChar (hd (String.explode str)))
				      orelse
				      let val cs = String.explode str
				      in (hd cs = #"<") 
					 andalso (List.last cs = #">") 
					 andalso (List.all isEnclosedChar (List.drop (rev (List.drop (cs,1)),1)))
				      end) alphabet
		then ()
		else raise Error ("alphabet should be a letter excluding " 
				  ^ (String.implode invalidAlphabetChars)
				  ^ " or string of alpha numerals and hyphen enclosed by < and >")
	val _ = if U.member "<EPSILON>" alphabet
		then raise Error "<EPSILON> should not be involved in alphabet"
		else ()
	val _ = if U.member "<EMPTY>" alphabet
		then raise Error "<EMPTY> should not be involved in alphabet"
		else ()

	fun prRule ((q,l),rs) = "   (" ^ q ^ "," ^ l ^ ") -> " ^ (U.toStringSpace (fn x=>x) rs)
	fun checkTrans ((q,l),qs) =
	    let val _ = case L.find (fn q => not (U.member q states)) qs of
			    NONE => ()
			  | SOME q => raise Error ("the range of transition function "
						   ^ (prRule ((q,l),qs)) ^ " contains invalid state " ^ q)
		val _ = case U.duplicatingElem qs of
			    NONE => ()
			  | SOME q => raise Error ("the range of transition function "
						   ^ (prRule ((q,l),qs)) ^ "contains duplicating state" ^ q)
	    in () end
	val _ = L.app checkTrans transitions
	val _ = let val dom = L.map (fn (x,y) => x) transitions
		in case U.duplicatingElem dom of
		    SOME (q,x) => raise Error ("transition for (" ^ q ^ "," ^ x ^ ") is multiply specified")
		  | NONE => ()
		end
    in true
    end
    
fun getDfa (states,alphabet,transitions,start,accept) =
    let val _ = case L.find (fn ((q,x),qs) => length qs <> 1)  transitions of
		    SOME ((q,x),_) => raise Error ("the image of transition function is multiple states for (" ^ q ^"," ^ x ^ ")")
		  | NONE => ()
	val dom1  = ListXProd.mapX (fn pair=>pair) (states,alphabet)
	val dom2 = L.map (fn (x,y) => x) transitions
	val _ =  case L.find (fn pair => not (U.member pair dom2)) dom1 of
		     NONE => ()
		   | SOME (q,x) => raise Error ("transition is not specified for (" ^ q ^"," ^ x ^ ")")
	val _ =  case L.find (fn pair => not (U.member pair dom1)) dom2 of
		     NONE => ()
		   | SOME (q,x) => raise Error ("invalid transition for (" ^ q ^"," ^ x ^ ")")
    in (states,alphabet,L.map (fn ((q,x),qs) => ((q,x),hd qs))transitions, hd start, accept)
    end

(* NFA では，transition の range が空集合のは記述しない(ファイルには記述してあっても良いが最初に削除する) *)
fun getNfa (states,alphabet,transitions,start,accept) =
    let val alphaEpsilon = alphabet @ ["<EPSILON>"]
	val dom = L.map (fn (x,y) => x) transitions
	val _ = case L.find (fn (q,x) => not (U.member q states) orelse not (U.member x alphaEpsilon)) dom of
		    NONE => ()
		  | SOME (q,x) => raise Error ("invalid transition for (" ^ q ^"," ^ x ^ ")")
    in (states,alphabet,L.filter (fn ((q,x),qs) => not (null qs)) transitions, hd start, accept)
    end

fun alpha (_,alphabet,_,_,_) = alphabet


(* Let key="--prefix=". If args contains "--prefix=str",  retrun  (SOME str, rest) *)
fun analyseOption key args = 
    let val (keyop,rest) = L.partition (String.isPrefix key) args
    in if null keyop
       then (NONE,rest)
       else let val opt = hd keyop
		val value = S.substring (opt, S.size key, (S.size opt) - (S.size key))
	    in (SOME value, rest)
	    end
    end

fun stripDquote str = let val cs = String.explode str
		      in if length cs > 2 andalso (hd cs) = #"\"" andalso (L.last cs) = #"\""
			 then S.substring (str, 1, (S.size str) - 2)
			 else str
		      end

fun analyseRenameOption args = 
    let val (prefix, rest1) = case analyseOption "--prefix=" args of
				 (NONE, rest) => ("q", rest)
			       | (SOME str, rest) => (stripDquote str,rest)
	val (numstr, rest2) = case analyseOption "--index=" rest1 of
				 (NONE, rest) => ("1", rest)
			       | (SOME str, rest) => (stripDquote str,rest)
    in (prefix,numstr,rest2)
    end

structure RegExpLrVals = RegExpLrValsFun (structure Token = LrParser.Token);
structure RegExpLex = RegExpLexFun (structure Tokens = RegExpLrVals.Tokens);
structure RegExpParser = Join (structure ParserData = RegExpLrVals.ParserData;
                             structure Lex = RegExpLex;
                             structure LrParser = LrParser);
fun readRegExp str =
    let val istream =TextIO.openString str
	val lexer = RegExpParser.makeLexer (fn n=>TextIO.inputN (istream, n));
	fun error (msg,line,col) = raise Error (msg ^ " at line " ^ (Int.toString line) ^ ", char "^ Int.toString col)
    in #1 (RegExpParser.parse (15, lexer, error, ()))
    end
    handle Io => raise Error ("failed to read the string " ^ str)

fun alphaInRegExp expr = L.filter (fn str => (str <> "<EPSILON>") andalso (str <> "<EMPTY>")) (Reg.symbols expr)

fun main (name, []) = (print helpMsg; OS.Process.failure)
  | main (name, opr::args) = 
    (case opr of
	"--dfa-check" => (if length args > 1
			 then (print helpMsg; OS.Process.failure)
			 else let val spec = if length args = 1
					     then compileFile (hd args)
					     else compileStdIn ()
				  val M = if isValidSpec spec then getDfa spec else raise Error ""
			      in (print "YES\n"; OS.Process.success)
			      end
			      handle Error msg => (print "NO\n"; printError msg; OS.Process.success))
     |  "--dfa-accept" => let val (opverbose,rest) = List.partition (fn str=> str = "--verbose") args
			      val verbose = not (null opverbose)
			  in if length rest > 2 orelse length rest < 1
			     then (print helpMsg; OS.Process.failure)
			     else let val cs = readInput (hd rest)
				      val spec = if length rest = 2
						 then compileFile (hd (tl rest))
						 else compileStdIn ()
				      val M = if isValidSpec spec then getDfa spec else raise Error ""
				      val _ = if L.all (fn s => U.member s (alpha M)) cs then ()
					      else raise Error "the alphabet of input string is invalid"
				      val _ = if Dfa.accept cs M then print "YES\n" else print "NO\n"
				      val _ = if verbose then Dfa.travel cs M else ()
			       in OS.Process.success
			       end
			  end
     | "--dfa-complement" => if length args > 1
			     then (print helpMsg; OS.Process.failure)
			     else let val spec = if length args = 1
						 then compileFile (hd args)
						 else compileStdIn ()
				      val M = if isValidSpec spec then getDfa spec else raise Error ""
				      val M' = Dfa.complement M
				  in (print (Dfa.toString M') ; OS.Process.success)
				  end								   
     | "--dfa-union" => if length args > 2 orelse length args < 1
			then (print helpMsg; OS.Process.failure)
			else let val spec1 = compileFile (hd args)
				 val spec2 = if length args = 2
					     then compileFile (hd (tl args))
					     else compileStdIn ()
				 val M1 = (if isValidSpec spec1 then getDfa spec1 else raise Error "")
					  handle Error msg => raise Error ("first DFA: " ^ msg)
				 val M2 = (if isValidSpec spec2 then getDfa spec2 else raise Error "")
					  handle Error msg => raise Error ("second DFA: " ^ msg)
				   val _ = if U.seteq (alpha M1, alpha M2)
					   then ()
					   else raise Error "two DFA should have the same alphabet"
				 val M = Dfa.union (M1,M2)
			     in (print (Dfa.toString M) ; OS.Process.success)
			     end								   
     | "--dfa-product" => if length args > 2 orelse length args < 1
			  then (print helpMsg; OS.Process.failure)
			  else let val spec1 = compileFile (hd args)
				   val spec2 = if length args = 2
					       then compileFile (hd (tl args))
					       else compileStdIn ()
				   val M1 = (if isValidSpec spec1 then getDfa spec1 else raise Error "")
					    handle Error msg => raise Error ("first DFA: " ^ msg)
				   val M2 = (if isValidSpec spec2 then getDfa spec2 else raise Error "")
					    handle Error msg => raise Error ("second DFA: " ^ msg)
				   val _ = if U.seteq (alpha M1, alpha M2)
					   then ()
					   else raise Error "two DFA should have the same alphabet"
				   val M = Dfa.intersection (M1,M2)
			     in (print (Dfa.toString M) ; OS.Process.success)
			     end								   
     | "--dfa-empty" => if length args > 1 then (print helpMsg; OS.Process.failure)
			else let val spec = if length args = 1
					    then compileFile (hd args)
					    else compileStdIn ()
				 val M = if isValidSpec spec then getDfa spec else raise Error ""
				 val _ = if Dfa.empty M then print "YES\n" else print "NO\n"
			     in OS.Process.success
			     end								   
     | "--dfa-subset" => if length args > 2 orelse length args < 1
			 then (print helpMsg; OS.Process.failure)
			 else let val spec1 = compileFile (hd args)
				  val spec2 = if length args = 2
					      then compileFile (hd (tl args))
					      else compileStdIn ()
				  val M1 = (if isValidSpec spec1 then getDfa spec1 else raise Error "")
					   handle Error msg => raise Error ("first DFA: " ^ msg)
				  val M2 = (if isValidSpec spec2 then getDfa spec2 else raise Error "")
					   handle Error msg => raise Error ("second DFA: " ^ msg)
				   val _ = if U.seteq (alpha M1, alpha M2)
					   then ()
					   else raise Error "two DFA should have the same alphabet"
				  val _ = if Dfa.subset (M1,M2) then print "YES\n" else print "NO\n"
			       in OS.Process.success
			       end								   
     | "--dfa-equiv" => if length args > 2 orelse length args < 1
			then (print helpMsg; OS.Process.failure)
			  else let val spec1 = compileFile (hd args)
				   val spec2 = if length args = 2
					      then compileFile (hd (tl args))
					      else compileStdIn ()
				   val M1 = (if isValidSpec spec1 then getDfa spec1 else raise Error "")
					    handle Error msg => raise Error ("first DFA: " ^ msg)
				   val M2 = (if isValidSpec spec2 then getDfa spec2 else raise Error "")
					    handle Error msg => raise Error ("second DFA: " ^ msg)
				   val _ = if U.seteq (alpha M1, alpha M2)
					   then ()
					   else raise Error "two DFA should have the same alphabet"
				   val _ = if Dfa.equivalent (M1,M2) then print "YES\n" else print "NO\n"
			       in OS.Process.success
			       end								   
     | "--dfa-minimize" => if length args > 1 then (print helpMsg; OS.Process.failure)
			   else let val spec = if length args = 1
					       then compileFile (hd args)
					       else compileStdIn ()
				    val M = if isValidSpec spec then getDfa spec else raise Error ""
				    val M' = Dfa.minimize M
				  in (print (Dfa.toString M') ; OS.Process.success)
				end
    | "--nfa-check" => (if length args > 1 then (print helpMsg; OS.Process.failure)
			else let val spec = if length args = 1
						then compileFile (hd args)
						else compileStdIn ()
				     val N = if isValidSpec spec then getNfa spec else raise Error ""
				 in (print "YES\n"; OS.Process.success)
				 end
			     handle Error msg => (print "NO\n"; printError msg; OS.Process.success))
    | "--nfa-accept" => let val (opverbose,rest) = List.partition (fn str=> str = "--verbose") args
			    val verbose = not (null opverbose)
			in if length rest > 2 orelse length rest < 1
			   then (print helpMsg; OS.Process.failure)
			   else let val cs = readInput (hd rest)
				    val spec = if length rest = 2
					       then compileFile (hd (tl rest))
					       else compileStdIn ()
				    val N = if isValidSpec spec then getNfa spec else raise Error ""
				    val alphabet  = alpha N
				    val _ = if L.all (fn s => U.member s alphabet) cs then ()
					    else raise Error "the alphabet of input string is invalid"
				    val _ = if Nfa.accept cs N then print "YES\n" else print "NO\n"
				    val _ = if verbose then Nfa.travel cs N else ()
				    val _ = if verbose then Nfa.showTransitionToAccept cs N else ()
				in OS.Process.success
			     end
			end
    | "--nfa-concat" => if length args > 2 orelse length args < 1
			then (print helpMsg; OS.Process.failure)
			  else let val spec1 = compileFile (hd args)
				   val spec2 = if length args = 2
					       then compileFile (hd (tl args))
					       else compileStdIn ()
				   val N1 = (if isValidSpec spec1 then getNfa spec1 else raise Error "")
					    handle Error msg => raise Error ("first NFA: " ^ msg)
				   val N2 = (if isValidSpec spec2 then getNfa spec2 else raise Error "")
					    handle Error msg => raise Error ("second NFA: " ^ msg)
				   val _ = if U.seteq (alpha N1, alpha N2)
					   then ()
					   else raise Error "two NFA should have the same alphabet"
				   val N = Nfa.concat (N1,N2)
			     in (print (Nfa.toString N) ; OS.Process.success)
			     end								   
    | "--nfa-union" => if length args > 2 orelse length args < 1
		       then (print helpMsg; OS.Process.failure)
			  else let val spec1 = compileFile (hd args)
				   val spec2 = if length args = 2
					       then compileFile (hd (tl args))
					       else compileStdIn ()
				   val N1 = (if isValidSpec spec1 then getNfa spec1 else raise Error "")
					    handle Error msg => raise Error ("first NFA: " ^ msg)
				   val N2 = (if isValidSpec spec2 then getNfa spec2 else raise Error "")
					    handle Error msg => raise Error ("second NFA: " ^ msg)
				   val _ = if U.seteq (alpha N1, alpha N2)
					   then ()
					   else raise Error "two NFA should have the same alphabet"
				   val N = Nfa.union (N1,N2)
			     in (print (Nfa.toString N) ; OS.Process.success)
			     end								   
    | "--nfa-star" => if length args > 1 then (print helpMsg; OS.Process.failure)
		      else let val spec = if length args = 1
					  then compileFile (hd args)
					  else compileStdIn ()
			       val N = if isValidSpec spec then getNfa spec else raise Error ""
			       val N' = Nfa.star N
			   in (print (Nfa.toString N') ; OS.Process.success)
			   end								   
    | "--nfa-collapse" => if length args > 1 then (print helpMsg; OS.Process.failure)
			  else let val spec = if length args = 1
					      then compileFile (hd args)
					  else compileStdIn ()
			       val N = if isValidSpec spec then getNfa spec else raise Error ""
			       val N' = Nfa.collapse N
			   in (print (Nfa.toString N') ; OS.Process.success)
			   end								   
    | "--nfa-mkdfa" => if length args > 1 then (print helpMsg; OS.Process.failure)
		      else let val spec = if length args = 1
					  then compileFile (hd args)
					  else compileStdIn ()
			       val N = if isValidSpec spec then getNfa spec else raise Error ""
			       val M = Nfa.powerset N
			   in (print (Dfa.toString M) ; OS.Process.success)
			   end								   
    | "--rename" => let val (prefix,numstr,rest) = analyseRenameOption args
			val num = case Int.fromString numstr of
				      NONE => raise Error "specify a number for --index=num"
				    | SOME num => num
		    in if length rest > 1
		       then (print helpMsg; OS.Process.failure)
		       else let val spec = if null rest then compileStdIn () else compileFile (hd rest)
				val N = if isValidSpec spec then getNfa spec else raise Error ""
				val N' = Nfa.rename (prefix,num) N
			    in (print (Nfa.toString N') ; OS.Process.success)
			    end
		    end
    | "--re-simplify" => if length args > 1 then (print helpMsg; OS.Process.failure)
			 else let val str = if length args = 1
					    then hd args
					    else case TextIO.inputLine TextIO.stdIn of
						     NONE => raise Error "no input of string"
						   | SOME str0 => str0
				  val expr = Reg.simplify (readRegExp str)
			      in (print ("\"" ^ (Reg.toString expr) ^ "\"\n") ; OS.Process.success)
			      end								   
    | "--re-mknfa" => if length args > 2 orelse length args < 1 
		      then (print helpMsg; OS.Process.failure)
		    else let val Sigma = readInput (hd args)
			     val _ = case U.duplicatingElem Sigma of
					 NONE => ()
				       | SOME str => raise Error "the alphabet contains dulicating elements"
			     val str = if length args = 2
				       then hd (tl args)
				       else case TextIO.inputLine TextIO.stdIn of
						NONE => raise Error "no input of string"
					      | SOME str0 => str0
			     val expr = readRegExp str
			     val _ = if L.all (fn s => U.member s Sigma) (alphaInRegExp expr) 
				     then ()
				     else raise Error "regexpr contains non-alphabet"
			     val N = Reg.expr2nfa Sigma expr
			 in (print (Nfa.toString N); OS.Process.success)
			 end								   
    | "--dfa-mkre" => let val (opverbose,rest) = List.partition (fn str=> str = "--verbose") args
			    val verbose = not (null opverbose)
		      in if length rest > 1 then (print helpMsg; OS.Process.failure)
			 else let val spec = if length rest = 1
					     then compileFile (hd rest)
					     else compileStdIn ()
				  val M = if isValidSpec spec then getDfa spec else raise Error ""
				  val str = Reg.dfa2re verbose M
				  val re = Reg.simplify (readRegExp str)
			      in (print ("\"" ^ (Reg.toString re) ^ "\"\n") ; OS.Process.success)
			      end			
		      end					   
    | "--dfa-pump" => if length args > 2 orelse length args < 1 
		      then (print helpMsg; OS.Process.failure)
		      else let val spec = if length args = 2
					then compileFile (hd (tl args))
					else compileStdIn ()
			     val M = if isValidSpec spec then getDfa spec else raise Error ""
			     val M' = Dfa.minimize M
			     val (_,alphabet,delta,q0,F) = M'
			     val (cs0,rest) = U.separateBy (fn x => x = #".") (String.explode (hd args))
			     val (cs1,cs2) = U.separateBy (fn x => x = #".") rest
			     val ss0 = readInput (String.implode cs0)
			     val ss1 = readInput (String.implode cs1)
			     val ss2 = readInput (String.implode cs2)
			     val _ = if L.all (fn s => U.member s alphabet) (ss0 @ ss1 @ ss2)
				     then ()
				     else raise Error "the alphabet of input string is invalid"
			     val _ = if null ss1
				     then raise Error "the pumping part is empty"
				     else ()
			     val q1 = Dfa.reach delta (q0, ss0)
			     val q2 = Dfa.reach delta (q1, ss1)
			     val q3 = Dfa.reach delta (q1, ss2)
			     val _ = if q1 = q2 andalso U.member q3 F
				     then print "YES\n" else print "NO\n"
			 in OS.Process.success
			 end								   
    | _ => (print helpMsg; OS.Process.failure)
    ) handle Error msg => (TextIO.output (TextIO.stdErr, "Error: " ^ msg ^ "\n"); OS.Process.failure)

end (* of local *)
end (* of struct *)
