(* file: util.sml *)
(* description: utility functions *)
(* author: Takahito Aoto *)

signature UTIL = 
sig 
    val mapAppend: ('a -> 'b list) -> 'a list -> 'b list
    val lookup: ''a -> (''a * 'b) list -> 'b option

    val member: ''a  -> ''a list -> bool
    val subseteq: ''a  list *  ''a list -> bool
    val union: ''a  list *  ''a list -> ''a list
    val mapUnion: ('a -> ''b list) -> 'a list -> ''b list
    val intersection: ''a  list *  ''a list -> ''a list
    val disjoint: ''a  list *  ''a list -> bool
    val seteq: ''a  list *  ''a list -> bool
    val setminus: ''a  list *  ''a list -> ''a list
    val distinct: ''a list -> bool
    val removeDuplication: ''a  list -> ''a list
    val powerset:  'a  list -> 'a list list
    val separateBy: ('a -> bool) -> 'a list -> 'a list * 'a list

    val orderedPairs: 'a list -> ('a * 'a) list

    val duplicatingElem: ''a list -> ''a option
    val takeUntilWithRest : ('a -> bool) -> 'a list -> 'a list * 'a list

    val toStringCommaRound: ('a -> string) -> 'a list -> string
    val toStringCommaBrace: ('a -> string) -> 'a list -> string
    val toStringSpace: ('a -> string) -> 'a list -> string
    val toStringCommaLnSquare: ('a -> string) -> 'a list -> string
end

structure Util : UTIL =
struct 

local 
    structure L = List
in

fun mapAppend f [] =  []
  | mapAppend f (x::xs) =  (f x) @  (mapAppend f xs)

fun lookup x [] = NONE
  | lookup x ((k,v)::ys) = if x = k then SOME v else lookup x ys

fun member x ys = List.exists (fn y => x = y) ys

fun subseteq (xs,ys) = List.all (fn x => member x ys) xs

fun rmDup [] = []
  | rmDup (x::xs) = if member x xs then rmDup xs else x :: (rmDup xs)

fun union (xs,ys) = rmDup (xs @ ys)
fun intersection (xs,ys) = List.filter (fn x => member x ys) xs
fun disjoint (xs,ys) = List.all (fn x=> not (member x ys)) xs
fun seteq (xs,ys) = subseteq (xs,ys) andalso subseteq (ys,xs)
fun setminus (xs,ys) = List.filter (fn x=> not (member x ys)) xs
fun distinct xs = (xs = rmDup xs)
fun removeDuplication [] = []
  | removeDuplication (x::xs) = if member x xs then removeDuplication xs
                                else x :: (removeDuplication xs)


fun mapUnion f [] = [] 
  | mapUnion f (x::xs) = union (f x, mapUnion f xs)

fun powerset xs = 
    let fun power [] =  [[]]
	  | power (n::ns) =  let val mss = power ns
			     in mss @ (List.map (fn ms=>n::ms) mss)
			     end
    in power (rev xs)
    end

fun orderedPairs [] = []
  | orderedPairs (x::ys) = (L.map (fn y => (x,y)) ys) @ orderedPairs ys


fun duplicatingElem [] = NONE
  | duplicatingElem (x::xs) = if member x xs then SOME x else duplicatingElem xs

fun separateBy p [] = ([],[])
  | separateBy p (c::cs) = if p c then ([],cs)
			   else let val (us,vs) = separateBy p cs
				in (c::us,vs)
				end


fun toStringBuilt prElm (start,sep,stop) xs =
    let fun toStringSub [] = ""
	  | toStringSub [x] = (prElm x)
	  | toStringSub (x::xs)= (prElm x) ^ sep ^ (toStringSub xs)
    in  start ^ (toStringSub xs) ^ stop
    end

fun takeUntilWithRest pred [] = ([],[])
  | takeUntilWithRest pred (x::xs) = 
    if pred x then ([],x::xs)
    else (fn (ys,zs) => (x::ys,zs)) (takeUntilWithRest pred xs)


fun toStringCommaRound prElm xs = toStringBuilt prElm  ("(",  ",",  ")") xs
fun toStringCommaBrace prElm xs = toStringBuilt prElm  ("{",  ",",  "}") xs
fun toStringSpace prElm xs = toStringBuilt prElm ("",  " ", "") xs
fun toStringCommaLnSquare prElm xs = toStringBuilt prElm ("   [ ",  ",\n     ",  " ]\n") xs


end (* of local *)

end
